import xml.etree.ElementPath

import attr


@attr.s
class Element:
    tag = attr.ib()
    attrib = attr.ib(default=())
    children = attr.ib(default=attr.Factory(list))

    def append(self, element):
        self.children.append(element)

    def extend(self, elements):
        self.children.extend(elements)

    def insert(self, index, element):
        self.children.insert(index, element)

    def remove(self, element):
        self.children.remove(element)

    def findall(self, path):
        return list(self.iterfind(path))

    def find(self, path):
        return next(self.iterfind(path), None)

    def iterfind(self, path):
        queryable_element = QueryableElement(self)
        found = xml.etree.ElementPath.iterfind(queryable_element, path)
        for item in found:
            if isinstance(item, ListBackedElement):
                yield item.list
            elif isinstance(item, QueryableElement):
                yield item.element

    def iter(self, tag=None):
        return iterate_element(self, tag)

    def __iter__(self):
        return iter(self.children)

    def __getitem__(self, index):
        return self.children[index]

    def __setitem__(self, index, element):
        self.children[index] = element

    def __len__(self):
        return len(self.children)


@attr.s
class QueryableElement:
    element = attr.ib()

    @property
    def tag(self):
        return self.element.tag

    def iter(self, tag=None):
        return iterate_element(self, tag)

    def __iter__(self):
        for item in self.element:
            if isinstance(item, Element):
                yield QueryableElement(item)
            elif isinstance(item, list):
                yield ListBackedElement(item)

    def __len__(self):
        return len(self.element)


@attr.s
class ListBackedElement:
    list = attr.ib()

    @property
    def tag(self):
        return self.list[0]

    def iter(self, tag=None):
        if tag is None or self.tag == tag:
            yield self

    def __iter__(self):
        return iter(())

    def __len__(self):
        return 0


def iterate_element(element, tag):
    if tag is None or element.tag == tag:
        yield element
    if not len(element):
        return
    for item in element:
        if hasattr(item, 'iter'):
            for subitem in item.iter(tag):
                yield subitem
        elif tag is None:
            yield item



"""RPP is a format used to describe `REAPER <http://reaper.fm>`_
projects."""

from .element import Element
from .rpp import dump, dumps, loads, load


__version__ = '0.3'
__author__ = 'Sviatoslav Abakumov <dust.harvesting@gmail.com>'
__all__ = ['dump', 'dumps', 'load', 'loads', 'Element']



import attr


def tokenize(string):
    lex = Lexer()
    lex.input(string)
    return iter(lex)


def lexer():
    return Lexer()


class Lexer:
    _input = None
    _iter = None

    def input(self, s):
        self._input = s
        self._iter = iter(self)

    def token(self):
        return next(self._iter, None)

    def __iter__(self):
        lines = self._input.splitlines()
        for lineno, line in enumerate(lines, start=1):
            is_first_token_in_line = True
            while line:
                line = line.strip()
                if starts_with_quote(line):
                    quote_end = self._find_closing_quote(line, lineno)
                    yield LexToken('STRING', line[1:quote_end], lineno)
                    line = line[quote_end+1:]
                else:
                    if is_first_token_in_line:
                        if line.startswith(OPEN):
                            yield LexToken('OPEN', OPEN, lineno)
                            line = line[1:]
                        elif line.startswith(CLOSE):
                            yield LexToken('CLOSE', CLOSE, lineno)
                            line = line[1:]
                        elif starts_with_pipe(line):
                            yield LexToken('STRING', line, lineno)
                            line = ''
                    if line:
                        pair = line.split(maxsplit=1)
                        thing, rest = pair if len(pair) > 1 else (pair[0], '')
                        yield LexToken('STRING', thing, lineno)
                        line = rest
                is_first_token_in_line = False
            yield LexToken('NEWLINE', '\n', lineno)

    def _find_closing_quote(self, line, lineno):
        try:
            return line.index(line[0], 1)
        except ValueError:
            raise ValueError('closing quote not found at line {}'.format(lineno))


@attr.s
class LexToken:
    type = attr.ib()
    value = attr.ib()
    lineno = attr.ib()


def starts_with_quote(s):
    quotes = '"\'`'
    return s[0] in quotes


def starts_with_pipe(s):
    return s[0] == '|'


tokens = (
    'OPEN',
    'CLOSE',
    'STRING',
    'NEWLINE',
)

OPEN = '<'
CLOSE = '>'



from setuptools import setup

with open('README.rst') as fp:
    README = fp.read()

setup(
    name='rpp',
    version='0.3',
    description='REAPER Project File Parser',
    long_description=README,
    author='Sviatoslav Abakumov',
    author_email='dust.harvesting@gmail.com',
    license='BSD',
    url='https://github.com/Perlence/rpp',
    download_url='https://github.com/Perlence/rpp/archive/0.3.tar.gz',
    packages=['rpp'],
    zip_safe=False,
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
    install_requires=[
        'ply',
        'attrs',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Text Processing',
    ])



from ply import yacc

from .element import Element
from .scanner import tokens  # noqa


def parser():
    return yacc.yacc(optimize=True, debug=False, write_tables=False)


def p_tree(t):
    """tree : OPEN root CLOSE NEWLINE
            | OPEN root items CLOSE NEWLINE"""
    t[0] = t[2]
    if len(t) > 5:
        t[0].extend(t[3])


def p_root(t):
    """root : list NEWLINE"""
    t[0] = Element(t[1][0], children=[])
    if len(t) > 2:
        t[0].attrib = t[1][1:]


def p_items(t):
    """items : item
             | item items"""
    if t[0] is None:
        t[0] = []
    t[0].append(t[1])
    if len(t) > 2:
        t[0] += t[2]


def p_item_list(t):
    """item : list NEWLINE"""
    if len(t[1]) == 1:
        t[0] = t[1][0]
    else:
        t[0] = t[1]


def p_item_tree(t):
    """item : tree"""
    t[0] = t[1]


def p_list(t):
    """list : STRING
            | STRING list"""
    if t[0] is None:
        t[0] = []
    t[0].append(t[1])
    if len(t) > 2:
        t[0] += t[2]


def p_error(t):
    if t is None:
        message = 'syntax error at EOF'
    else:
        message = 'syntax error at line {}, token={}'.format(t.lineno or 1, t.type)
    raise ValueError(message)



from collections import Iterable
from decimal import Decimal

from .element import Element
from .scanner import starts_with_quote, starts_with_pipe


def encode(element, indent=2, level=0):
    result = ' ' * level * indent
    if isinstance(element, Element):
        result += '<'
        result += encode_tag_and_attrib(element)
        for item in element:
            result += encode(item, level=level+1)
        result += ' ' * level * indent + '>\n'
    elif isinstance(element, str):
        result += quote_string(element, quote_pipe=False) + '\n'
    else:
        result += encode_value(element) + '\n'
    return result


def encode_tag_and_attrib(element):
    result = element.tag
    if element.attrib:
        result += ' ' + encode_iterable(element.attrib)
    result += '\n'
    return result


def encode_value(value):
    if isinstance(value, str):
        return quote_string(value)
    elif isinstance(value, Iterable):
        return encode_iterable(value)
    elif isinstance(value, Decimal):
        return quote_string(format(value, 'f'))
    else:
        return quote_string(str(value))


def encode_iterable(iterable):
    return ' '.join(map(encode_value, iterable))


def quote_string(value, quote_pipe=True):
    if not value:
        return '""'
    if not should_quote(value, quote_pipe):
        return value
    quote, value = quote_mark(value)
    return '{quote}{value}{quote}'.format(quote=quote, value=value)


def should_quote(s, quote_pipe):
    return (quote_pipe or not starts_with_pipe(s)) and (starts_with_quote(s) or has_whitespace(s))


def has_whitespace(s):
    whitespace = ' \t'
    return any(ch in whitespace for ch in s)


def quote_mark(s):
    quote = '"'
    if '"' in s:
        quote = "'"
    if "'" in s:
        quote = '`'
    if '`' in s:
        quote = '`'
        s = s.replace('`', "'")
    return quote, s



from . import scanner
from .decoder import parser
from .encoder import encode


def loads(string):
    lexer = scanner.lexer()
    yacc = parser()
    return yacc.parse(string, lexer)


def load(fp):
    return loads(fp.read())


def dumps(lists, indent=2):
    return encode(lists, indent=indent)


def dump(lists, fp, indent=2):
    fp.write(dumps(lists, indent))
